import os
import json
import docker
import docker.auth

from sbdk.error import SbdkError
from .runner import Runner, ImageRunner
from .shell import Shell, Bash, SSH
from .utils import find_image, parse_repository_tag
from sbgsdk.util import import_name


def make_client(project, **kwargs):
    if project.context.get('docker_client_class'):
        cls = import_name(project.context.get('docker_client_class'))
        return cls(project, **kwargs)
    timeout = 60
    if os.path.exists('/home/sbg/.sbg/internal/config.json'):
        with open('/home/sbg/.sbg/internal/config.json') as conf:
            cfg = json.load(conf)
            timeout = cfg.get('docker_daemon_timeout', 60)
    args = {'base_url': project.docker_daemon_url,
            'version': project.docker_protocol_version,
            'timeout': timeout}
    args.update(kwargs)
    docker_client = docker.Client(**args)
    return docker_client


def make_runner(project, **kwargs):
    if project.project_data:
        return Runner(make_client(project), project, **kwargs)
    return ImageRunner(make_client(project), project, **kwargs)


def make_shell(project, **kwargs):
    kind = project.shell_type
    if kind == 'bash':
        args = {'docker': make_runner(project)}
        args.update(kwargs)
        return Bash(project, **args)
    elif kind == 'ssh':
        return Bash(project, **kwargs)
    else:
        raise SbdkError("Unknown shell: '%s'" % kind)




import os
import logging

import sbdk.resources as resources
from sbdk.commands.fix_uid import fix_uid


def argument_parser(subparsers):
    """ Setups parser for init subcommand """

    parser = subparsers.add_parser("init", help="Initialize project by populating project directory")
    parser.add_argument("project_name", type=str, default='',
                        help="Project name. It will be used as Python package name as well.")

    parser.add_argument("--base-image", "-b", type=str, default=None,
                        help="Base Docker image for new project")

    parser.add_argument("--example", "-e", default=False, action='store_true',
                        help="If this option is provided, project will be "
                             "populated with example working wrapper")
    parser.set_defaults(func=init)


def init(project, project_name, base_image, example):
    project.initialize(project_name, base_image)
    mkdirs(project, example)
    mkfiles(project, example)
    fix_uid(project, str(os.getuid()))
    project.state


def mkdirs(project, example):
    dst = project.project_dir
    top_level = os.path.join(dst, project.name)
    os.makedirs(top_level)
    os.makedirs(os.path.join(dst, 'test-data'))
    if example:
        os.makedirs(os.path.join(top_level, 'grep', 'test'))


def mkfiles(project, example):
    dst = project.project_dir

    resources.process_template('setup.py.tpl', os.path.join(dst, 'setup.py'), project=project)
    resources.copy('logging.json', os.path.join(dst, '.sbdk', 'logging.json'))
    with open(os.path.join(dst, project.name, '__init__.py'), 'w'):
        pass

    if example:
        resources.process_template('init.py.tpl', os.path.join(dst, project.name, '__init__.py'), project=project)

        packages = ['grep']
        file_names = '__init__.py', 'wrapper.py', 'schema.py', 'test.py'
        files = [(dir_name, file_name) for dir_name in packages for file_name in file_names]
        for f in files:
            dist_path = os.path.join(dst, project.name, f[0], f[1])
            resources.process_template('example-project/top_level/%s/%s.tpl' % f,
                                       dist_path,
                                       project=project)
        resources.copy('example-project/top_level/grep/test/words.txt',
                       os.path.join(dst, project.name, 'grep', 'test', 'words.txt'))



__author__ = 'sinisa'

from sbgsdk import define
import logging

class OutputWrapper(define.Wrapper):
    for_each = ['inputA', 'params.group_by']

    class Inputs(define.Inputs):
        input = define.input(
            list=True,
            name="input",
            description="",
            file_types=[],
        )

    class Params(define.Params):
        path = define.string(
            name = "Output directory",
            description = "Output directory.",
        )

    def execute(self):
        for input in self.inputs.input:
            logging.info(input)


from sbgsdk import schema as _
from wrapper import Wrapper

Inputs = _.Inputs
input = _.IOAttr

Outputs = _.Outputs
output = _.IOAttr

Params = _.Params
integer = _.IntAttr
boolean = _.BoolAttr
enum = _.EnumAttr
real = _.RealAttr
string = _.StringAttr
struct = _.StructAttr
__wrapper__ = Wrapper



## Mapping of exceptions to integer error codes.
## make sure all values are DISTINCT integers in range (32..127]

NOT_A_JOB_ERROR = 33

JOB_ERROR = 34

NO_WRAPPER_ID_ERROR = 36

INPUTS_VALIDATION_ERROR = 37

PARAMS_VALIDATION_ERROR = 38

EXIT_CODE_MAPPING_ERROR = 39

NO_SUCH_METHOD_ERROR = 40

NO_RESOURCES_ERROR = 41

INTERNAL_ERROR = 101


import uuid
from sbgsdk import define
from sbgsdk.spw import SimpleParallelWrapper
from sbgsdk.tests.test_tools.comparator import comparator


class BlueComparator(SimpleParallelWrapper):
    for_each = ['inputA', 'params.group_by']

    class Inputs(define.Inputs):
        inputA = define.input(
            list=True,
            name="hex file A",
            description="Text file containing a list of hex characters",
            file_types=['text'],
        )

    class Outputs(define.Outputs):
        output = define.output(
            list=True,
            name="output",
            description="Text file containing a list of hex characters",
            file_types=['text'],
        )

    class Params(define.Params):
        group_by = define.enum(['sample', 'library', 'platform_unit', 'chunk', None])

    def execute(self):
        inputs, outputs = self.inputs, self.outputs
        filename = str(uuid.uuid4()) + '.txt'
        assert len(inputs.inputA) == 2, 'input len = %s' % len(inputs.inputA)
        comparator.compare(inputs.inputA[0], inputs.inputA[1], filename)
        outputs.output.add_file(filename).meta = inputs.inputA.make_metadata(file_type='text')


import sys

from sbdk.commands import check_image
from sbdk.docker import make_shell


def argument_parser(subparsers):
    parser = subparsers.add_parser("sh", help="Run shell in your image")
    parser.set_defaults(func=sh)

    parser.add_argument("--command", "-c", help="Command to execute", type=str)


def sh(project, command):
        check_image(project)
        shl = make_shell(project)

        if command:
            container = shl.run_command(command)
        else:
            container = shl.run_shell()

        ask_commit(container, project.state)
        container.remove()


def ask_commit(container, save):
    sys.stdout.write("Do you want to commit this container? [Y/n]: ")
    while True:
        choice = sys.stdin.readline().lower().strip()
        if choice == '' or choice[0] == 'y':
            container.commit('', tag=False)
            if save:
                container.save_as_base()
            else:
                sys.stdout.write("Image id: {}".format(container.image['Id']))
            return
        elif choice == 'n':
            return
        else:
            sys.stdout.write("Please enter 'y' or 'n': ")



from sbgsdk.schema.io_value import IOValue
from sbgsdk.util import DotAccessDict, intersect_dicts


class IOList(object):
    def __init__(self, value=None):
        value = value or []
        if isinstance(value, basestring):
            value = [value]
        self._values = [IOValue(v) for v in value or []]

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        return iter(self._values)

    def __getitem__(self, i):
        return self._values[i]

    def __json__(self):
        return self._values

    __str__ = __unicode__ = __repr__ = lambda self: str(self._values)

    def add_file(self, path=None):
        """ Add new file to list of outputs and return the object. """
        value = IOValue(path)
        self._values.append(value)
        return value

    def make_metadata(self, **kwargs):
        """ Intersect metadata dictionaries of each file in list and returns the new Metadata object. """
        if not self._values:
            return DotAccessDict(**kwargs)
        if len(self._values) == 1:
            return self[0].make_metadata(**kwargs)
        return DotAccessDict(reduce(intersect_dicts, [v.meta for v in self._values]), **kwargs)

    def _save_meta(self):
        for val in self:
            val._save_meta()

    def _load_meta(self):
        for val in self:
            val._load_meta()



import json


def argument_parser(subparsers):
    parser = subparsers.add_parser("show-settings", help="Show project configuration")
    parser.set_defaults(func=show)


def show(project):
    print(json.dumps(project.context, indent=2))



from sbgsdk.schema.io_classes import IOAttr, SchemaBased, IODef, SchemaBased
from sbgsdk.schema.io_params import BoolAttr, EnumAttr, IntAttr, RealAttr, StringAttr, StructAttr

Input = Output = IOAttr
Params = SchemaBased
Inputs = Outputs = IODef


from sbgsdk.errors import *


class JobException(Exception):
    def __json__(self):
        return {'$$type': 'error', 'message': self.message or str(self)}

    @classmethod
    def from_dict(cls, d):
        return cls(d.get('message', 'Job error'))


################################################# PROTOCOL EXCEPTIONS ##################################################
class ProtocolException(JobException):
    pass


class NoWrapperIdException(ProtocolException):
    pass


class NoResourcesException(ProtocolException):
    pass


class NoSuchMethodError(ProtocolException):
    pass


class NotAJobError(ProtocolException):
    pass


################################################# VALIDATION EXCEPTIONS ################################################
class ValidationException(JobException):
    pass


class InputsValidationException(ValidationException):
    pass


class ParamsValidationException(ValidationException):
    pass


########################################################################################################################
__exit_codes = {
    InputsValidationException: INPUTS_VALIDATION_ERROR,
    ParamsValidationException: PARAMS_VALIDATION_ERROR,
    JobException: JOB_ERROR,
    NoWrapperIdException: NO_WRAPPER_ID_ERROR,
    NotAJobError: NOT_A_JOB_ERROR,
    NoSuchMethodError: NO_SUCH_METHOD_ERROR,
    NoResourcesException: NO_RESOURCES_ERROR
}


def exit_code_for_exception(exception):
    return __exit_codes.get(exception.__class__, EXIT_CODE_MAPPING_ERROR)



import os

from sbgsdk import define
from sbgsdk.tests.test_tools.generator import generator
from sbgsdk.sgw import ScatterGatherWrapper


class GeneratorWrapper(ScatterGatherWrapper):
    class Outputs(define.Outputs):
        output = define.output(
            list=True,
            name="output",
            description="Text file containing a list of hex characters",
            #file_types=['text'],
        )

    class Params(define.Params):
        number_of_files = define.integer(required=True, description='How many files to generate, Master?', default=1)
        numbers_per_file = define.integer(required=True, description='How many numbers per file, Master?', default=1024)
        template_number = define.integer(required=True, description='What number to write in files, Master?', default=0)
        numbers_to_generate = define.enum(
            required=True,
            values=['same', 'random'],
            name="Numbers To Generate",
            description="Type of numbers to generate (same or random). [default: random]",
            default="random",
        )

    def execute(self):
        outputs, params = self.outputs, self.params
        for i in range(0, params.number_of_files):
            filename = 'hex' + str(i) + '.txt'
            generator.generate(filename, params.numbers_per_file, params.numbers_to_generate, params.template_number)
            outputs.output.add_file(filename).meta = outputs.output.make_metadata(file_type='text', sample='s'+str(i))

    def split(self):
        return [{'job_index': x} for x in xrange(self.params.number_of_files)]

    def work(self, job):
        assert 'job_index' in job, str(job)
        p = self.params
        file_path = os.path.abspath('hex' + str(job['job_index']) + '.txt')
        generator.generate(file_path, p.numbers_per_file, p.numbers_to_generate, p.template_number)
        return file_path

    def merge(self, job_results):
        for result in job_results:
            out = self.outputs.output.add_file(result)
            out.meta = {"file_type": 'text', "sample": os.path.basename(result)}



from sbdk.commands import check_project_dir, check_project_data, check_project_state
from sbdk.auth import update_docker_cfg
from docker.auth.auth import load_config

def argument_parser(subparsers):
    parser = subparsers.add_parser("logout", help="Login to docker registry")
    parser.set_defaults(func=registry_logout)

def registry_logout(project):
    check_project_dir(project)
    check_project_data(project)
    check_project_state(project)

    cfg = load_config()
    if project.docker_registry_url in cfg:
        del cfg[project.docker_registry_url]

    update_docker_cfg(cfg)




import json
import os.path
import requests
import logging
from sys import exit

from sbdk.auth import push_url, push_json
from sbdk.commands import check_project_dir, check_image, check_project_data
from sbdk.docker import make_runner
from sbdk.error import SbdkError
from sbdk.docker.container import dockerize


def argument_parser(subparsers):
    parser = subparsers.add_parser("push", help="Create new image and upload it")
    parser.set_defaults(func=push)
    parser.add_argument('message', help='Short description of your changes')


def push(project, message, retries=2, container=None):
    if retries < 0:
        container.remove()
        raise SbdkError('Push command failed')
    check_project_dir(project)
    check_project_data(project)
    check_image(project)

    if not container:
        docker = make_runner(project)
        container = docker.run_install()
        if not container.is_success():
            print("Error while preparing image:")
            container.print_log()
            exit("Push command failed")
    config = {'Cmd': project.sdk_runner_command, 'Entrypoint': project.docker_entrypoint}
    container.load_credentials(renew=False)
    if not container.check_privilege():
        print('You do not have permissions to push into repository. Try again:')
        container.load_credentials(renew=True)
        if not container.check_privilege():
            container.remove()
            raise SbdkError('You do not have permissions to push into repository')
    container.commit(message, config).push()
    if not check_container_push(container):
        retry = raw_input('Push failed, image not found in registry. Try again? [Y/n]: ').lower().strip()
        if retry == '' or retry == 'y':
            push(project, message, retries-1, container)
        elif retry == 'n':
            container.remove()
            exit("Push command failed")
    push_schema(project, container)
    container.remove()


def push_schema(project, container):
    data_dir = project.project_data
    # nulling project_data to force image runner
    project.project_data = None
    project.image_id = container.image['Id']
    runner = make_runner(project)
    runner.schema('.sbdk/schema.json', remove=True)

    with open(os.path.join(data_dir, 'schema.json')) as fp:
        wrapper_data = json.load(fp)  # json.loads(self.docker.logs(c))

    if not wrapper_data:
        print('No wrappers registered (empty __init__.py?). Exiting.')
        exit(1)

    push_data = {'message': container.message or 'NO COMMIT MESSAGE',
                 'image_data': {'id': container.repository,
                                'tag': container.tag},
                 'wrappers': wrapper_data}

    if project.schema_server_url:
        url = push_url(project, dockerize(project.state['username']))
        try:
            push_json(url, push_data, container.token)
        except requests.HTTPError as e:
            if e.response.status_code in (401, 402):
                container.load_credentials()
                push_json(url, push_data, container.token)
            else:
                raise e


def check_container_push(container, retries=2):
    if retries < 0:
        return False
    docker_repo, image_id = container.repository, container.tag

    check_image = requests.get('%s://%s/v1/repositories/%s/%s/images'
                               % tuple([container.project.docker_registry_schema] + docker_repo.split('/')))
    check_tag = requests.get('%s://%s/v1/repositories/%s/%s/tags'
                             % tuple([container.project.docker_registry_schema] + docker_repo.split('/')))
    if check_image.status_code != 200:
        raise Exception('Bad response (%s) : %s' % (check_image.status_code, check_image.content))
    if check_tag.status_code != 200:
        raise Exception('Bad response (%s) : %s' % (check_tag.status_code, check_tag.content))
    if any(image_id in tag for tag in [i['id'] for i in check_image.json()]):
        if image_id in check_tag.json().keys():
            logging.debug('Image pushed successfuly')
            return True
        else:
            logging.debug('Image pushed but not taged. Trying again: Retry left: %s' % str(retries))
            container.push()
            return check_container_push(container, retries-1)




import random


def generate(filename, count=1024, numbers_to_generate='random', template_number=0):
    """ outputs a list of hex numbers into a file (in range 0x00 to 0xFF)"""
    count = int(count)
    if count < 1:
        raise ValueError("count cannot be negative, count=%s" % str(count))

    with open(filename, 'w') as f:
        f.write(str(count) + '\n')

        if numbers_to_generate == 'random':
            for i in xrange(count):
                number = hex(random.randint(0, 255))
                f.write(number+'\n')
        elif numbers_to_generate == 'same':
            input_number = int(template_number)
            if input_number != input_number % 256:
                raise ValueError("invalid template_number: %s" % template_number)
            number = hex(input_number)
            for i in xrange(count):
                f.write(number+'\n')
        else:
            raise ValueError("unknown numbers_to_generate=%s" % numbers_to_generate)




from sbdk.commands import check_project_state
from subprocess import Popen
from sys import stdout


def argument_parser(subparsers):
    parser = subparsers.add_parser("pull", help="Fetch docker images from repository")
    parser.set_defaults(func=pull)


def pull(project):
    check_project_state(project)
    pull = Popen(['docker', 'pull', project.repo_name], stdout=stdout)
    pull.wait()



from nose.tools import *

import sbdk.docker.runner as runner


def test_shell_join():
    result = runner.shell_join(['a', 'b'])
    eq_(result, 'a b', 'Bad simple escape: %s' % result)

    result = runner.shell_join(['a', 'b c'])
    eq_(result, "a 'b c'", 'Bad whitespace escape: %s' % result)

    result = runner.shell_join(['a', 'b"c'])
    eq_(result, "a 'b\"c'", 'Bad double quote escape: %s' % result)

    #result = self.runner.shell_escape('a', "b'c")
    #eq_(result, "a b\\'c", 'Bad single quote escape: %s' % result)


def test_multi_command():
    eq_(runner.multi_command('a', 'b'), ['/bin/sh', '-c', 'a && b'], 'Bad simple join')



import tempfile
import os

from sbgsdk.exceptions import InputsValidationException, ParamsValidationException, NoSuchMethodError
from sbgsdk.schema import SchemaBased, IODef
from sbgsdk.job import Job, Resources, Outputs
from sbgsdk.executor import Tassadar


class Wrapper(object):
    class Inputs(IODef):
        pass

    class Outputs(IODef):
        pass

    class Params(SchemaBased):
        pass

    def __init__(self, inputs, params, context=None, resources=None):
        self.inputs = inputs if isinstance(inputs, IODef) else self.Inputs(**inputs)
        self.params = params if isinstance(params, SchemaBased) else self.Params(**params)
        self.outputs = self.Outputs()
        self.context = context or {}
        self.resources = resources or Resources()
        self.inputs._load_meta()

    def execute(self):
        pass

    def __call__(self, method, args):
        method = method or '_entry_point'
        if not callable(getattr(self, method, None)):
            raise NoSuchMethodError('Wrapper has no method %s' % method)
        result = getattr(self, method)(**args)
        if result is None:
            self.outputs._save_meta()
            out_dict = self.outputs.__json__()
            for item in self.outputs:
                if item.id not in out_dict or not item.value:
                    out_dict[item.id] = []
            return Outputs(out_dict)
        return result

    def get_requirements(self):
        return _get_method_requirements(self, 'execute') or Resources()

    def _entry_point(self):
        errors = self.inputs._validate()
        if errors:
            raise InputsValidationException(unicode(self.__class__)+u'\n'+unicode(errors))
        errors = self.params._validate()
        if errors:
            raise ParamsValidationException(unicode(errors))
        return self.job('execute', requirements=self.get_requirements())

    @classmethod
    def _get_schema(cls):
        return {
            'inputs': cls.Inputs._get_schema(),
            'outputs': cls.Outputs._get_schema(),
            'params': cls.Params._get_schema(),
        }

    def job(self, method=None, args=None, requirements=None, name=None):
        return Job(self.__class__, method=method,
                   resources=requirements or _get_method_requirements(self, method) or Resources(), args=args,
                   inputs=self.inputs.__json__(), params=self.params.__json__(), context=self.context)

    def test(self):
        test_exec_dir = tempfile.mkdtemp(prefix='test_%s_' % self.__class__.__name__, dir='.')
        os.chdir(test_exec_dir)
        try:
            self.inputs._validate(assert_=True)
            self.params._validate(assert_=True)
            self.inputs._save_meta()
            outputs = Tassadar().execute(self.job()).outputs
            self.outputs = self.Outputs(**outputs)
            self.outputs._load_meta()
            self.outputs._validate(assert_=True)
        finally:
            os.chdir('..')
        return self.outputs

    def get_allocated_memory(self, units='MB'):
        units = units.upper()
        converter = {
            'B': lambda b: b,
            'KB': lambda b: b / 1024,
            'MB': lambda b: b / 1024**2,
            'GB': lambda b: b / 1024**3,
        }
        if units not in converter:
            raise ValueError('Units argument must be one of: %s' % converter.keys())
        return converter[units](self.resources.mem_mb * 1024**2)


def _get_method_requirements(wrapper, method_name):
    if not method_name:
        return getattr(wrapper, '_requirements', None)
    m = getattr(wrapper, method_name, None)
    return getattr(m, '_requirements', None) or getattr(wrapper, '_requirements', None)


from sbgsdk.test import assert_files_identical
from sbgsdk.tests.common import cleanup
from sbgsdk.tests.test_tools.generator import generator
from sbgsdk.tests.test_wrappers import BlueComparator





@cleanup
def test_comparator_sample():
    w = BlueComparator(
        inputs={'inputA': ['hex0.txt', 'hex1.txt', 'hex2.txt', 'hex3.txt']},
        params={'group_by': 'sample'},
    )

    w.inputs.inputA[0].meta.file_type = 'text'
    w.inputs.inputA[0].meta.sample = 'S1'
    w.inputs.inputA[1].meta.file_type = 'text'
    w.inputs.inputA[1].meta.sample = 'S1'
    w.inputs.inputA[2].meta.file_type = 'text'
    w.inputs.inputA[2].meta.sample = 'S2'
    w.inputs.inputA[3].meta.file_type = 'text'
    w.inputs.inputA[3].meta.sample = 'S2'
    w.inputs._save_meta()

    count = 1000
    generator.generate('hex0.txt', count, 'same', 15)
    generator.generate('hex1.txt', count, 'same', 15)
    generator.generate('hex2.txt', count, 'same', 42)
    generator.generate('hex3.txt', count, 'same', 42)
    w.test()

    s1 = filter(lambda x: x.meta.sample == 'S1', w.outputs.output)[0]
    s2 = filter(lambda x: x.meta.sample == 'S2', w.outputs.output)[0]
    assert_files_identical(s1, 'hex0.txt')
    assert_files_identical(s2, 'hex2.txt')


@cleanup
def test_comparator_platform_unit():
    w = BlueComparator(
        inputs={'inputA': ['hex0.txt', 'hex1.txt', 'hex2.txt', 'hex3.txt']},
        params={'group_by': 'platform_unit'},
    )

    w.inputs.inputA[0].meta.file_type = 'text'
    w.inputs.inputA[0].meta.sample = 'S1'
    w.inputs.inputA[0].meta.platform_unit = 'PU1'
    w.inputs.inputA[1].meta.file_type = 'text'
    w.inputs.inputA[1].meta.sample = 'S1'
    w.inputs.inputA[1].meta.platform_unit = 'PU1'
    w.inputs.inputA[2].meta.file_type = 'text'
    w.inputs.inputA[2].meta.sample = 'S1'
    w.inputs.inputA[2].meta.platform_unit = 'PU2'
    w.inputs.inputA[3].meta.file_type = 'text'
    w.inputs.inputA[3].meta.sample = 'S1'
    w.inputs.inputA[3].meta.platform_unit = 'PU2'
    w.inputs._save_meta()

    count = 1000
    generator.generate('hex0.txt', count, 'same', 15)
    generator.generate('hex1.txt', count, 'same', 15)
    generator.generate('hex2.txt', count, 'same', 42)
    generator.generate('hex3.txt', count, 'same', 42)
    w.test()

    s1 = filter(lambda x: x.meta.platform_unit == 'PU1', w.outputs.output)[0]
    s2 = filter(lambda x: x.meta.platform_unit == 'PU2', w.outputs.output)[0]
    assert_files_identical(s1, 'hex0.txt')
    assert_files_identical(s2, 'hex2.txt')


@cleanup
def test_comparator_no_group():
    w = BlueComparator(
        inputs={'inputA': ['hex0.txt', 'hex1.txt']},
        params={'group_by': None},
    )

    w.inputs.inputA[0].meta.file_type = 'text'
    w.inputs.inputA[0].meta.sample = 'S1'
    w.inputs.inputA[0].meta.platform_unit = 'PU1'
    w.inputs.inputA[1].meta.file_type = 'text'
    w.inputs.inputA[1].meta.sample = 'S1'
    w.inputs.inputA[1].meta.platform_unit = 'PU1'
    w.inputs._save_meta()

    count = 1000
    generator.generate('hex0.txt', count, 'same', 15)
    generator.generate('hex1.txt', count, 'same', 15)
    w.test()

    assert_files_identical(w.outputs.output[0], 'hex0.txt')



#!/usr/bin/env python

import sys
import argparse
import logging

from sbdk.commands import init, push, pull, test, sh, run, show_settings, schema, fix_uid, upload, login, logout
from sbdk.error import SbdkError
from sbdk.project import Project
from sbdk import VERSION


def main(argv=None, context=None):

    context = context or {}
    parser = argparse.ArgumentParser(description="Seven Bridges Wrapper SDK Tool",
                                     epilog="Try 'sbdk [COMMAND] -h' for command specific options",
                                     prog="sbg",
                                     version=VERSION)
    parser.add_argument('--project-dir', default='',
                        help='Use this dir instead of project dir.')
    parser.add_argument('--verbose', '-V', default=False, action='store_true',
                        help='Increase output verbosity')

    subparsers = parser.add_subparsers()
    commands = [init, test, push, pull, sh, run, show_settings, schema, fix_uid, upload, login, logout]
    for cmd in commands:
        cmd.argument_parser(subparsers)

    args = vars(parser.parse_args(argv))

    level = logging.DEBUG if args.pop('verbose') else logging.INFO
    logging.basicConfig(stream=sys.stderr, level=level)

    context['project_dir'] = args.pop('project_dir')
    project = Project(context)

    f = args.pop('func')
    underscored = dict([(k.replace('-', '_'), v) for k, v in args.items()])
    try:
        f(project, **underscored)
    except SbdkError as e:
        print(e.message)
        sys.exit(1)


if __name__ == '__main__':
    main(sys.argv[1:])



import os
import nose.tools
from sbgsdk.tests.test_tools.generator import generator
from sbgsdk.tests.test_tools.adder import adder


def remove_txt_files():
    for name in os.listdir('.'):
        if name.endswith('.txt'):
            os.remove(name)


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
def test_add_same():
    count = 1000
    generator.generate('same_a.txt', count, 'same', 22)
    generator.generate('same_b.txt', count, 'same', 33)
    adder.add('same_a.txt', 'same_b.txt', 'same_c.txt')

    f = open('same_c.txt')
    first_line = f.readline()
    count2 = int(first_line.rstrip())
    nose.tools.assert_equals(count, count2)

    line_count = 0
    for line in f:
        integer_num = int(line.rstrip(), 16)
        assert(integer_num == 55)
        line_count += 1
    nose.tools.assert_equals(line_count, count)

    f.close()


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
@nose.tools.raises(Exception)
def test_add_diff_length():
    generator.generate('same_a.txt', 1000, 'same', 45)
    generator.generate('same_b.txt', 1001, 'same', 45)
    adder.add('same_a.txt', 'same_b.txt', 'same_c.txt')


def test_split_deltas():
    res = adder._split_deltas(12, 3)
    assert res[0] == 4
    assert res[1] == 4
    assert res[2] == 4

    res = adder._split_deltas(11, 3)
    assert res[0] == 4
    assert res[1] == 4
    assert res[2] == 3

    res = adder._split_deltas(10, 3)
    assert res[0] == 4
    assert res[1] == 4
    assert res[2] == 2

    res = adder._split_deltas(9, 3)
    assert res[0] == 3
    assert res[1] == 3
    assert res[2] == 3

    res = adder._split_deltas(3, 3)
    assert res[0] == 1
    assert res[1] == 1
    assert res[2] == 1


@nose.tools.raises(ValueError)
def test_split_deltas_e1():
    adder._split_deltas(12, 0)


@nose.tools.raises(ValueError)
def test_split_deltas_e2():
    adder._split_deltas(12, -1)


@nose.tools.raises(ValueError)
def test_split_deltas_e3():
    adder._split_deltas(12, 13)


@nose.tools.raises(ValueError)
def test_split_deltas_e4():
    adder._split_deltas(0, 0)


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
def test_split():
    generator.generate('numbers.txt', 11, 'random')
    adder.split('numbers.txt', 3)
    assert_size(adder.get_part_name('numbers.txt', 0), 4)
    assert_size(adder.get_part_name('numbers.txt', 1), 4)
    assert_size(adder.get_part_name('numbers.txt', 2), 3)


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
def test_unsplit():
    generator.generate('numbers.txt', 11, 'random')
    file_names = adder.split('numbers.txt', 3)
    adder.unsplit('out.txt', file_names)
    assert_size('out.txt', 11)


def assert_size(filename, n):
    fa = open(filename, 'rb+')
    total_count = int(fa.readline())
    assert total_count == n
    fa.close()




import copy
from .attrdef import AttrDef


class SchemaBased(object):
    def __init__(self, **kwargs):
        for attr, attr_def in self._attr_defs().iteritems():
            setattr(self, attr, attr_def(attr, kwargs.get(attr)))

    @classmethod
    def _attr_defs(cls):
        result = {}
        for key in dir(cls):
            val = getattr(cls, key)
            if isinstance(val, AttrDef):
                result[key] = val
        return result

    @classmethod
    def _get_schema(cls):
        schema = []
        attr_defs = cls._attr_defs().items()
        for attr, attr_def in sorted(attr_defs, cmp=lambda x,y: cmp(x[1]._order, y[1]._order)):
            attr_def(attr) # Set default name and desc if none supplied.
            attr_schema = dict(attr_def.get_schema(), id=attr)
            schema.append(attr_schema)
        return schema

    def _validate(self, assert_=False):
        errors = {}
        for attr, attr_def in self._attr_defs().iteritems():
            try:
                attr_def.validate(getattr(self, attr))
            except (ValueError, TypeError) as e:
                errors[attr] = str(e)
        if assert_:
            # noinspection PyStringFormat
            assert not errors, 'Following errors encountered: %s' % errors
        return errors

    def _definitions(self):
        result = {}
        for id, attr in self.__class__._attr_defs().iteritems():
            new = copy.deepcopy(attr)
            new.value = getattr(self, id)
            new.id = id
            result[id] = new
        return result

    def __json__(self):
        return dict([(k, v) for k, v in self.__dict__.iteritems() if k[0] != '_' and v is not None])
    __unicode__ = __repr__ = __str__ = lambda self: str(self.__json__())

    def __iter__(self):
        for k, v in self._definitions().iteritems():
            yield v

    def __getitem__(self, item):
        return self.__json__()[item]


import os

def add(filename_a, filename_b, filename_out):
    """ compares two hex list files and if same -> outputs same file, else error"""
    with open(filename_a, 'r') as fa, open(filename_b, 'r') as fb:
        count_a = int(fa.readline())
        count_b = int(fb.readline())
        assert count_a == count_b, "Different sizes: count_a(%d) != count_b(%d)" % (count_a, count_b)
        with open(filename_out, 'w') as fc:
            fc.write(str(count_a)+'\n')
            for i in xrange(count_a):
                a = int(fa.readline(), 16)
                b = int(fb.readline(), 16)
                number = hex((a+b) % 256)
                fc.write(number+'\n')


def _split_deltas(count, n):
    if count <= 0:
        raise ValueError("count must be positive! count=%s" % count)
    if n <= 0:
        raise ValueError("n must be positive! n=%s" % n)
    if count < n:
        raise ValueError("count cannot be less than n! count=%s, n=%s" % (count, n))

    delta = count / n
    if count % n != 0:
        delta += 1

    lst = []
    while count > 0:
        if delta < count:
            lst.append(delta)
            count -= delta
        else:
            lst.append(count)
            count = 0
    return lst


def get_part_name(filename, i):
    base, ext = os.path.splitext(filename)
    return base + "_" + str(i) + ext


def split(filename, n):
    """ split files into n new files """

    with open(filename, 'r') as fa:
        total_count = int(fa.readline())
        part_counts = _split_deltas(total_count, n)
        file_names = []
        for i, part_count in enumerate(part_counts):
            part_name = get_part_name(filename, i)
            with open(part_name, 'w') as file_part:
                file_names.append(part_name)
                file_part.write(str(part_count) + '\n')
                for j in xrange(part_count):
                    file_part.write(fa.readline())
    return file_names


def unsplit(outfile, file_names):
    """merge list of files into new file"""

    total_size = 0
    # get total size from first lines
    for file_name in file_names:
        with open(file_name, 'r') as fh:
            total_size += int(fh.readline())

    # write numbers from partial files into outfile
    with open(outfile, 'w') as out:
        out.write(str(total_size) + '\n')  # write total num of lines
        for file_name in file_names:
            with open(file_name, 'r') as fh:
                size = int(fh.readline())
                for x in xrange(size):
                    out.write(fh.readline())

    return outfile




import os
import nose
from nose.tools import assert_equals, raises
from sbgsdk.tests.test_tools.generator import generator


def remove_txt_files():
    for name in os.listdir('.'):
        if name.endswith('.txt'):
            os.remove(name)


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
def test_generator_random():
    count = 1000
    generator.generate('random.txt', count, 'random')
    f = open('random.txt')
    first_line = f.readline()
    count2 = int(first_line.rstrip())
    assert_equals(count, count2)

    line_count = 0
    for line in f:
        integer_num = int(line.rstrip(), 16)
        assert(0 <= integer_num < 256)
        line_count += 1
    assert_equals(line_count,count)

    f.close()


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
def test_generator_ones():
    count = 1000
    generator.generate('ones.txt', count, 'same', 1)
    f = open('ones.txt')
    first_line = f.readline()
    count2 = int(first_line.rstrip())
    assert_equals(count, count2)

    line_count = 0
    for line in f:
        num_ones = int(line.rstrip(), 16)
        assert(num_ones == 1)
        line_count += 1
    assert_equals(line_count, count)

    f.close()


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
def test_generator_zeros():
    count = 1000
    generator.generate('zeros.txt', count, 'same', 0)
    f = open('zeros.txt')
    first_line = f.readline()
    count2 = int(first_line.rstrip())
    assert_equals(count, count2)

    line_count = 0
    for line in f:
        num_ones = int(line.rstrip(), 16)
        assert(num_ones == 0)
        line_count += 1
    assert_equals(line_count, count)

    f.close()


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
@raises(ValueError)
def test_generator_wrong_count():
    count = -5
    generator.generate('zeros.txt', count, 'same', 0)


@nose.tools.with_setup(remove_txt_files, remove_txt_files)
@raises(ValueError)
def test_generator_wrong_type():
    count = 1000
    generator.generate('zeros.txt', count, 'batman')



__author__ = 'luka'


class MockClient(object):

    def __init__(self, *args, **kwargs):
        pass

    def attach(self, container):
        yield u"Mock output"

    def attach_socket(self, container, params=None, ws=False):
        pass

    def build(self, path=None, tag=None, quiet=False, fileobj=None,
              nocache=False, rm=False, stream=False):
        return u'0123456789ab', u"Mock message"

    def commit(self, container, repository=None, tag=None, message=None,
               author=None, conf=None):
        return {u'Id': u'0123456789ab'}

    def containers(self, quiet=False, all=False, trunc=True, latest=False,
                   since=None, before=None, limit=-1):
        return [{u'Command': u'/bin/bash ',
                 u'Created': 1390325161,
                 u'Id': u'2dcb4ff19477720b4cb1d04d80b11ccf6f3bce327ca5d37e265fcb4ba10a3485',
                 u'Image': u'ubuntu:12.04',
                 u'Names': [u'/kickass_albattani'],
                 u'Ports': None,
                 u'SizeRootFs': 0,
                 u'SizeRw': 0,
                 u'Status': u'Up 17 minutes'}]

    def copy(self, container, resource):
        return u"Mock raw result"

    def create_container(self, image, command=None, hostname=None, user=None,
                         detach=False, stdin_open=False, tty=False,
                         mem_limit=0, ports=None, environment=None, dns=None,
                         volumes=None, volumes_from=None, privileged=False,
                         name=None):
        return {u'Id': u'0123456789ab'}

    def create_container_from_config(self, config, name=None):
        return {u'Id': u'0123456789ab'}

    def diff(self, container):
        pass

    def events(self):
        yield {u'from': u'ubuntu:12.04',
               u'id': u'2dcb4ff19477720b4cb1d04d80b11ccf6f3bce327ca5d37e265fcb4ba10a3485',
               u'status': u'die',
               u'time': 1390326539}

    def export(self, container):
        return u"Mock raw result"

    def history(self, image):
        return u'[{"Id":"This json is left unparsed"}]'

    def images(self, name=None, quiet=False, all=False, viz=False):
        return [{u'Created': 1364102658,
                 u'Id': u'0123456789ab9a3c593ef05b4332b1d1a02a62b4accb2c21d589ff2f5f2dc',
                 u'Repository': u'ubuntu',
                 u'Size': 24653,
                 u'Tag': u'quantal',
                 u'VirtualSize': 180116135}]

    def import_image(self, src, data=None, repository=None, tag=None):
        pass

    def info(self):
        return {u'Containers': 25,
                u'Debug': 0,
                u'Driver': u'aufs',
                u'DriverStatus': [[u'Root Dir', u'/var/lib/docker/aufs'], [u'Dirs', u'65']],
                u'IPv4Forwarding': 1,
                u'Images': 11,
                u'IndexServerAddress': u'https://index.docker.io/v1/',
                u'InitPath': u'/usr/bin/docker',
                u'InitSha1': u'',
                u'KernelVersion': u'3.5.0-45-generic',
                u'LXCVersion': u'0.8.0-rc1',
                u'MemoryLimit': 1,
                u'NEventsListener': 3,
                u'NFd': 11,
                u'NGoroutines': 16,
                u'SwapLimit': 1}

    def insert(self, image, url, path):
        pass

    def inspect_container(self, container):
        return {u'Args': [],
                u'Config': {u'AttachStderr': True,
                            u'AttachStdin': True,
                            u'AttachStdout': True,
                            u'Cmd': [u'/bin/bash'],
                            u'CpuShares': 0,
                            u'Dns': None,
                            u'Domainname': u'',
                            u'Entrypoint': None,
                            u'Env': None,
                            u'ExposedPorts': {},
                            u'Hostname': u'2dcb4ff19477',
                            u'Image': u'ubuntu',
                            u'Memory': 0,
                            u'MemorySwap': 0,
                            u'NetworkDisabled': False,
                            u'OpenStdin': True,
                            u'PortSpecs': None,
                            u'StdinOnce': True,
                            u'Tty': True,
                            u'User': u'',
                            u'Volumes': {},
                            u'VolumesFrom': u'',
                            u'WorkingDir': u''},
                u'Created': u'2014-01-21T17:26:01.421055749Z',
                u'Driver': u'aufs',
                u'HostConfig': {u'Binds': None,
                                u'ContainerIDFile': u'',
                                u'Links': None,
                                u'LxcConf': [],
                                u'PortBindings': {},
                                u'Privileged': False,
                                u'PublishAllPorts': False},
                u'HostnamePath': u'/var/lib/docker/containers/2dcb4ff19477720b4cb1d04d80b11ccf6f3bce327ca5d37e265fcb4ba10a3485/hostname',
                u'HostsPath': u'/var/lib/docker/containers/2dcb4ff19477720b4cb1d04d80b11ccf6f3bce327ca5d37e265fcb4ba10a3485/hosts',
                u'ID': u'2dcb4ff19477720b4cb1d04d80b11ccf6f3bce327ca5d37e265fcb4ba10a3485',
                u'Image': u'8dbd9e392a964056420e5d58ca5cc376ef18e2de93b5cc90e868a1bbc8318c1c',
                u'Name': u'/kickass_albattani',
                u'NetworkSettings': {u'Bridge': u'',
                                     u'Gateway': u'',
                                     u'IPAddress': u'',
                                     u'IPPrefixLen': 0,
                                     u'PortMapping': None,
                                     u'Ports': None},
                u'Path': u'/bin/bash',
                u'ResolvConfPath': u'/var/lib/docker/containers/2dcb4ff19477720b4cb1d04d80b11ccf6f3bce327ca5d37e265fcb4ba10a3485/resolv.conf',
                u'State': {u'ExitCode': 0,
                           u'FinishedAt': u'2014-01-21T17:48:59.923387731Z',
                           u'Ghost': False,
                           u'Pid': 0,
                           u'Running': False,
                           u'StartedAt': u'2014-01-21T17:26:01.483899231Z'},
                           u'Volumes': {},
                           u'VolumesRW': {}}

    def inspect_image(self, image_id):
        return {u'Size': 522943178,
                u'architecture': u'x86_64',
                u'config': {u'AttachStderr': False,
                            u'AttachStdin': False,
                            u'AttachStdout': False,
                            u'Cmd': None,
                            u'CpuShares': 0,
                            u'Dns': None,
                            u'Domainname': u'',
                            u'Entrypoint': None,
                            u'Env': None,
                            u'ExposedPorts': None,
                            u'Hostname': u'',
                            u'Image': u'',
                            u'Memory': 0,
                            u'MemorySwap': 0,
                            u'NetworkDisabled': False,
                            u'OpenStdin': False,
                            u'PortSpecs': None,
                            u'StdinOnce': False,
                            u'Tty': False,
                            u'User': u'',
                            u'Volumes': None,
                            u'VolumesFrom': u'',
                            u'WorkingDir': u''},
                u'container': u'd61fc212316b7812410364b1c6e6859eb0d2ed47c49b821fe70fac3273d340b1',
                u'container_config': {u'AttachStderr': True,
                                      u'AttachStdin': True,
                                      u'AttachStdout': True,
                                      u'Cmd': [u'/bin/bash'],
                                      u'CpuShares': 0,
                                      u'Dns': [u'8.8.8.8', u'8.8.4.4'],
                                      u'Domainname': u'',
                                      u'Entrypoint': [],
                                      u'Env': None,
                                      u'ExposedPorts': {},
                                      u'Hostname': u'd61fc212316b',
                                      u'Image': u'ubuntu:12.04',
                                      u'Memory': 0,
                                      u'MemorySwap': 0,
                                      u'NetworkDisabled': False,
                                      u'OpenStdin': True,
                                      u'PortSpecs': None,
                                      u'StdinOnce': True,
                                      u'Tty': True,
                                      u'User': u'',
                                      u'Volumes': {},
                                      u'VolumesFrom': u'',
                                      u'WorkingDir': u''},
                u'created': u'2013-11-04T17:18:41.638987857+01:00',
                u'docker_version': u'0.6.5',
                u'id': u'9c7d47054d404c3f1273df8d6dd544937084fa0b4c4afbc79b8b2cc035417a2e',
                u'parent': u'8dbd9e392a964056420e5d58ca5cc376ef18e2de93b5cc90e868a1bbc8318c1c'}

    def kill(self, container, signal=None):
        pass

    def login(self, username, password=None, email=None, registry=None):
        return {'Status': 'Login Succeeded'}

    def logs(self, container):
        return '{"message":"logs"}'

    def port(self, container, private_port):
        return 10666

    def pull(self, repository, tag=None, stream=False):
        return u'{"status": "another unparsed json"}'

    def push(self, repository, stream=False):
        return u'{"status": "another unparsed json"}'

    def remove_container(self, container, v=False, link=False):
        pass

    def remove_image(self, image):
        pass

    def restart(self, container, timeout=10):
        pass

    def search(self, container, timeout=10):
        return []

    def start(self, container, binds=None, port_bindings=None, lxc_conf=None,
              publish_all_ports=False, links=None):
        pass

    def stop(self, container, timeout=10):
        pass

    def tag(self, image, repository, tag=None, force=False):
        return True

    def top(self, container):
        return {u'Processes': [[u'root',
                u'14810',
                u'14786',
                u'0',
                u'18:26',
                u'pts/5',
                u'00:00:00',
                u'/bin/bash']],
                u'Titles': [u'UID', u'PID', u'PPID', u'C', u'STIME', u'TTY', u'TIME', u'CMD']}

    def version(self):
        return {u'Arch': u'amd64',
                u'GitCommit': u'bc3b2ec',
                u'GoVersion': u'go1.2',
                u'KernelVersion': u'3.5.0-45-generic',
                u'Os': u'linux',
                u'Version': u'0.7.6'}

    def wait(self, container):
        return 0


class OutJson(MockClient):

    def create_container_from_config(self, config, name=None):
        if config['Cmd'][0] != 'chown':
            with open('__out__.json', 'w') as f:
                    f.write('{}')
        return {u'Id': u'0123456789ab'}


import os

from sbdk.error import SbdkError
from sbdk.docker.container import Container

class Shell(object):

    def run_shell(self):
        raise NotImplementedError(self._message('run_shell'))

    def run_command(self, command):
        raise NotImplementedError(self._message('run_command'))

    def _message(self, method):
        return "Method '%s' not implemented for class %s." % method, self.__class__.__name__


class Bash(Shell):

    def __init__(self, project, docker):
        self.project = project
        self.cid_file = os.path.join(self.project.project_data, 'cid')
        self.docker = docker

    def run_shell(self):
        pid = os.fork()
        if not pid:
            self.sh()
        else:
            os.wait()
            with open(self.cid_file) as cid_file:
                cid = cid_file.read()
            os.remove(self.cid_file)
            return Container(self.docker.client, self.project, {'Id': cid})

    def sh(self):
        args = ['docker', 'run', '-i', '-t']
        args += ['-v', '%s:/sbgenomics:rw' % self.project.project_dir]
        args += ['-v', '%s:/sb:rw' % self.project.sb_home]
        args += ['--cidfile', self.cid_file]
        args += [self.project.state['image_id']]
        args += self.project.docker_entrypoint
        args += ['/bin/bash']

        os.execvp('docker', args)

    def run_command(self, command):
        cont = self.docker.run(['/bin/bash', '-c', command])
        cont.print_log()
        return cont


class SSH(Shell):
    def run_shell(self):
        raise NotImplementedError(self._message('run_shell'))

    def run_command(self, command):
        raise NotImplementedError(self._message('run_command'))


import json


def classify(obj):
    """ Make a deep copy of obj that contains classes in place of {"$type": "<type>", ...} dicts """
    from sbgsdk.job import Job, Resources, Outputs
    from sbgsdk.exceptions import JobException
    type_map = {
        'job': Job,
        'resources': Resources,
        'outputs': Outputs,
        'error': JobException,
    }
    if isinstance(obj, (list, tuple)):
        return [classify(item) for item in obj]
    if isinstance(obj, dict):
        if '$$type' not in obj:
            return {k: classify(v) for k, v in obj.iteritems()}
        return type_map[obj.pop('$$type')].from_dict(obj)
    return obj


def from_json(str_or_fp, object_hook=None):
    """ Load json and make classes from certain dicts (see classify() docs) """
    if isinstance(str_or_fp, basestring):
        obj = json.loads(str_or_fp, object_hook=object_hook)
    else:
        obj = json.load(str_or_fp, object_hook=object_hook)
    return classify(obj)


def to_json(obj, fp=None):
    default = lambda o: o.__json__() if callable(getattr(o, '__json__', None)) else str(o)
    kwargs = dict(default=default, indent=2, sort_keys=True)
    return json.dump(obj, fp, **kwargs) if fp else json.dumps(obj, **kwargs)


import shutil


def compare(filename_a, filename_b, filename_out):
    """ compares two hex list files and if same -> outputs same file, else error"""

    with open(filename_a, 'r') as fa, open(filename_b, 'r') as fb:
        count_a = int(fa.readline())
        count_b = int(fb.readline())

        assert count_a == count_b, "Different sizes: count_a(%d) != count_b(%d)" % (count_a, count_b)

        for i in xrange(count_a):
            a = int(fa.readline(), 16)
            b = int(fb.readline(), 16)
            assert a == b, "Different numbers on line#%d: a=%d != b=%d" % (i, a, b)

    shutil.copyfile(filename_a, filename_out)



import os
import stat
import logging

from sbgsdk.job import Job
from sbgsdk.exceptions import JobException
from sbgsdk.protocol import from_json, to_json
from sbgsdk.executor import Executor


class Adun(Executor):
    def __init__(self, runner):
        self.runner = runner

    def execute(self, job, one_container=False, remove=True):
        logging.info('Job started: %s' % job.job_id)
        logging.debug('Job: %s' % to_json(job))

        if one_container:
            return self.exec_wrapper_full(job, remove=remove)

        job.resolved_args = self.resolve(job.args)
        job.status = Job.RUNNING
        result = self.exec_wrapper_job(job, remove=remove)
        if isinstance(result, JobException):
            raise result
        job.status = Job.DONE
        if isinstance(result, Job):
            return self.execute(result)
        logging.info('Job finished: %s' % to_json(job))
        return result

    def exec_wrapper_full(self, job, remove=True):
        job_dir = job.job_id
        os.mkdir(job_dir)
        os.chmod(job_dir, os.stat(job_dir).st_mode | stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH)
        os.chdir(job_dir)
        with open('__in__.json', 'w') as fp:
            logging.debug('Writing job order to %s', os.path.abspath('__in__.json'))
            to_json(job, fp)
        self.runner.run_wrapper('__in__.json',
                                cwd=job_dir, remove=remove)
        with open('__out__.json') as fp:
            logging.debug('Reading job output from %s', os.path.abspath('__out__.json'))
            result = from_json(fp)
        os.chdir('..')
        return result

    def exec_wrapper_job(self, job, remove=True):
        job_dir = job.job_id
        if not os.path.exists(job_dir):
            os.mkdir(job_dir)
        os.chmod(job_dir, os.stat(job_dir).st_mode | stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH)
        in_file, out_file = [os.path.join(job_dir, f) for f in '__in__.json', '__out__.json']
        with open(in_file, 'w') as fp:
            logging.debug('Writing job order to %s', in_file)
            to_json(job, fp)
        self.runner.run_job('__in__.json', '__out__.json', cwd=job_dir, remove=remove)
        with open(out_file) as fp:
            logging.debug('Reading job output from %s', out_file)
            result = from_json(fp)
        from subprocess import Popen
        Popen(['sudo chmod -R 777 ' + job_dir], shell=True)  # TODO: remove
        Popen(['sudo chown -R 1001:1001 ' + job_dir], shell=True)  # TODO: remove
        return result

    def resolve(self, val):
        if isinstance(val, Job):
            return self.execute(val)
        if isinstance(val, (list, tuple)):
            return [self.resolve(item) for item in val]
        if isinstance(val, dict):
            return {k: self.resolve(v) for k, v in val.iteritems()}
        return val



import os
import shutil
import nose


from sbgsdk import Process
from sbgsdk.tests.test_wrappers.red_adder import RedAdder

RED_ADDER_ID = '.'.join([RedAdder.__module__, RedAdder.__name__])


def remove_temp_files():
    prefixes = ['process_', 'random_a_', 'random_b_', 'output_']
    suffixes = ['.txt']
    temp_files = ['__out__.json', '__exec__.log']
    wrapper_names = 'BlueComparator', 'GeneratorWrapper', 'RedAdder'
    for filename in os.listdir('.'):
        if any(map(filename.startswith, prefixes)) or any(map(filename.endswith, suffixes)):
            temp_files.append(filename)
    for filename in temp_files:
        if os.path.isfile(filename):
            os.remove(filename)
    for dirname in os.listdir('.'):
        if os.path.isdir(dirname) and any(map(dirname.startswith, ['test_%s_' % wrp for wrp in wrapper_names])):
            shutil.rmtree(dirname)


cleanup = nose.tools.with_setup(remove_temp_files, remove_temp_files)


def run_cli(*args):
    p = Process('python', '-m', 'sbgsdk.cli', *args)
    p.run()
    return p




import json
import logging
import os
import time
import itertools
import functools
import select
from subprocess import Popen, PIPE


DEVNULL = open(os.devnull, 'w+b')


def print_prepare(arg):
    return json.dumps(arg) if ' ' in arg else arg


class SubprocessExecutionException(Exception):
    """ Raised by the Process class if subprocess finished with exit_code != 0
    User error message is constructed from cmd_line and stderr or stdout.
    """
    def __init__(self, proc, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)
        self.proc = proc


class _ProcOutput(object):
    def __init__(self, file_path=None, line_callback=None):
        self._w_handle = None
        self._r_handle = None
        self.mode = 'wb'
        self._unhandled_line_part = ''
        self.file_path = file_path
        self.line_callback = line_callback

    def get_lines(self):
        if not self.file_path:
            return []
        with open(self.file_path) as f:
            return f.readlines()

    def get_string(self):
        return ''.join(self.get_lines())

    def _get_popen_param(self):
        if self.file_path:
            self._w_handle = open(self.file_path, self.mode)
            if self.line_callback:
                self._r_handle = open(self.file_path)
            return self._w_handle
        return DEVNULL

    def _close(self):
        if self._w_handle:
            self._w_handle.close()
        if self._r_handle:
            self._r_handle.close()

    def _callback(self, ready_handle_list):
        if not self._r_handle in ready_handle_list:
            return
        lines = self._r_handle.read().splitlines(True)
        if not lines:
            return
        lines[0] = self._unhandled_line_part + lines[0]
        if not lines[-1].endswith(('\n', '\r', '\r\n')):
            self._unhandled_line_part = lines.pop()
        for line in lines:
            self.line_callback(line)


def _get_stdin_popen_param(proc):
    if not proc.stdin:
        return DEVNULL
    if isinstance(proc.stdin, basestring):
        return open(proc.stdin)
    if isinstance(proc.stdin, (list, tuple)):
        return PIPE
    raise Exception('Bad stdin: %s', proc.stdin)


class Process(object):
    """ Use this class to run external processes. Example:
    >>> p = Process('python', '-c', 'x,y=input(),input(); print x+y', stdin=['2', '2']).run()
    >>> p.stdout.get_string()
    4

    Passing positional arguments to __init__ is same as using add_arg(*args).

    kwargs: stdin
    >>> Process('wc', stdin='/path/to/file').run() # /path/to/file will be streamed to wc stdin
    >>> Process('wc', stdin=['first line', 'second line']) # pass list to stdin line by line

    kwargs: stdout and stderr
    >>> money_lines = []
    >>> Process('grep', r'\$\d+', stdin='file.txt', stdout=lambda line: money_lines.append(line)).run() # pass callable
    >>> Process('wc', 'file.txt', stderr='errors.log')

    You can also set these using process.[stdout|stderr].line_parser and process.[stdout|stderr].file_path
    """
    _ids = itertools.count(0)

    def __init__(self, *args, **kwargs):
        self.id = kwargs.get('id') or 'process_%s' % Process._ids.next()
        self.popen = None
        self.cwd = kwargs.get('cwd')
        self.args = []
        self.add_arg(*args)

        self.stdin = kwargs.get('stdin')

        self.stderr = _ProcOutput()
        stderr = kwargs.get('stderr', True)
        if stderr:
            self.stderr.file_path = stderr if isinstance(stderr, basestring) else '%s.err' % self.id
            if callable(stderr):
                self.stderr.line_callback = stderr

        self.stdout = _ProcOutput()
        stdout = kwargs.get('stdout', True)
        if stdout:
            self.stdout.file_path = stdout if isinstance(stdout, basestring) else '%s.out' % self.id
            if callable(stdout):
                self.stdout.line_callback = stdout

    def run(self, wait=True):
        """ Run the process. If wait is true, wait for it to finish and parse output if line parsers are supplied. """
        try:
            stdout = self.stdout._get_popen_param()
            stderr = self.stderr._get_popen_param()
            stdin = _get_stdin_popen_param(self)
            logging.info('Executing: %s\n\tcmd_line: %s\n\tstdin: %s\n\tstdout: %s\n\tstderr: %s\n',
                         self.id, self.cmd_line, unicode(stdin), unicode(stdout), unicode(stderr))
            self.popen = Popen(self.args, stdin=stdin, stdout=stdout, stderr=stderr, cwd=self.cwd)
            if stdin == PIPE:
                time.sleep(.1)
                for line in self.stdin:
                    line = line + '\n' if not line.endswith('\n') else line
                    self.popen.stdin.write(line)
                    self.popen.stdin.flush()
        except Exception:
            self._close()
            raise
        return self.wait() if wait else self

    def wait(self):
        """ Wait for the process to finish and parse output if line parsers are supplied. """
        try:
            rlist = []
            if self.stdout.line_callback:
                rlist.append(self.stdout._r_handle)
            if self.stderr.line_callback:
                rlist.append(self.stderr._r_handle)
            if not rlist:
                self.popen.wait()
            else:
                select_args = rlist, [], [], .1
                done = False
                while not done:
                    if self.popen.poll() is not None:
                        done = True
                    ready = select.select(*select_args)[0]
                    if ready:
                        self.stdout._callback(ready)
                        self.stderr._callback(ready)
            level = logging.WARN if self.exit_code else logging.INFO
            message = 'Process finished with exit code %s: %s' % (self.exit_code, self.cmd_line)
            logging.log(level, message)
            if self.exit_code:
                raise SubprocessExecutionException(self, message)
        finally:
            self._close()
        return self

    def _close(self):
        self.stdout._close()
        self.stderr._close()
        if self.popen and isinstance(self.popen.stdin, file):
            self.popen.stdin.close()

    def kill(self):
        self._close()
        self.popen.kill()

    @property
    def cmd_line(self):
        """ Get the command line that can be used to rerun the process. """
        return ' '.join(map(print_prepare, self.args))

    @property
    def exit_code(self):
        """ Exit code returned by the process. None if process not finished. """
        return None if self.popen is None else self.popen.poll()

    @property
    def is_running(self):
        return self.popen is not None and self.exit_code is None

    def add_arg(self, *args, **kwargs):
        """ Append arguments to the argument list. For supplied kwargs, add_narg is called. """
        if self.popen is not None:
            raise Exception('Cannot add arguments once process has been executed.')
        for arg in args:
            if arg not in (None, ''):
                self.args.append(unicode(arg).strip())
        for key, val in kwargs.iteritems():
            self.add_narg(key, val)

    def add_narg(self, key, value):
        """ Add named arguments to the argument list. Examples:
        >>> p = Process('python')
        >>> p.add_narg('c', 'print 2*2') # Same as p.add_arg('-c', 'print 2*2')
        >>> p.add_narg('3', True) # Same as p.add_arg('-3'). For bool values, only key is added or nothing if False.
        """
        if value in (None, '') or value is False:
            return
        if not key.startswith('-'):
            key = '-' + key if len(key) == 1 else '--' + key
        if isinstance(value, bool):
            self.add_arg(key)
        else:
            self.add_arg(key, value)


def process_fails(exit_code, str_to_search=""):
    """ decorator for test functions that call Process.run() """
    def decorator(func):
        @functools.wraps(func)
        def wrapped():
            try:
                func()
                assert False, "SubprocessExecutionException expected, but not raised!"
            except SubprocessExecutionException as e:
                assert exit_code == e.proc.exit_code, 'Wrong exit code. expected: %s, actual: %s' % (exit_code, e.proc.exit_code)
                assert str_to_search in e.proc.stderr.get_string(), 'string "%s" was not found.' % str_to_search
        return wrapped
    return decorator



import getpass
import json
import requests
import logging
import os
import getpass
import base64

from sbdk.error import SbdkError

DOCKER_CONFIG_FILENAME = '.dockercfg'

def login(project, retries=2):
    if retries < 0:
        raise SbdkError('Login failed')
    url = project.schema_server_url + '/auth'
    usr = raw_input("Username: ")
    if not usr:
        return login(project, retries-1)
    pwd = getpass.getpass()
    resp = push_json(url, {'username': usr, 'password': pwd}, None, no_raise_for=[401, 402])
    if resp.status_code in (401, 402):
        print('Sorry. Try again.')
        return login(project, retries-1)
    return usr, resp.json()['session_id']


def push_url(project, user):
    return project.schema_server_url + '/repo/' + user + '/' + project.name + '/image'


def push_json(url, data, session_id, no_raise_for=tuple()):
    headers = {'Content-Type': 'application/json'}
    if session_id:
        headers['session-id'] = session_id
    response = requests.post(url, json.dumps(data), headers=headers, verify=False)
    if response.status_code not in no_raise_for:
        response.raise_for_status()
    return response


def store_credentials(docker_registry_name, username, password, cfg, email=None):
    email = email or '%s@example.com' % username

    if not cfg.get(docker_registry_name, None):
        cfg[docker_registry_name] = {}
    cfg[docker_registry_name]['username'], cfg[docker_registry_name]['password'], cfg[docker_registry_name]['email'] \
        = username, password, email
    update_docker_cfg(cfg)


def update_docker_cfg(cfg):
    def convert_to_dockercfg(cfg):
        def encode_auth(username, password):
            return base64.b64encode(username + b':' + password)
        dockercfg = {}
        for repo, auth in cfg.iteritems():
            dockercfg[repo] = {'auth': encode_auth(auth['username'], auth['password']), 'email': auth['email']}
        return dockercfg

    config_file = os.path.join(os.environ.get('HOME', '.'), DOCKER_CONFIG_FILENAME)
    with open(config_file, 'w') as outfile:
        json.dump(convert_to_dockercfg(cfg), outfile)


from setuptools import setup, find_packages
import sys
import os.path

sys.path.append(os.path.dirname(__file__))

from sbdk import VERSION


with open('requirements-dev.txt') as f:
    required = f.read().splitlines()

setup(
    name="sbgsdk",
    version=VERSION,
    include_package_data=True,
    install_requires=required,
    packages=find_packages(),
    entry_points={
        'console_scripts': ['sbg = sbdk.main:main'],
    }
)



import json

from sbgsdk.job import Job
from sbdk.executor import Adun
from sbgsdk.protocol import to_json
from sbdk.docker import make_runner
from sbdk.commands import check_image, check_project_dir


def argument_parser(subparsers):
    parser = subparsers.add_parser("run", help="Run wrapper")
    parser.set_defaults(func=run)

    parser.add_argument("wrapper", help="Wrapper name", type=str)
    parser.add_argument("args", help="Load wrapper args from this JSON file", type=str)
    parser.add_argument("--output", "-o", help="Save wrapper output to this JSON file", type=str)
    parser.add_argument("--keep-container", "-k", help="Don't remove container when run finishes", type=bool)
    # parser.add_argument("--one-container", help="Run all the jobs from single container", type=bool)


def run(project, wrapper, args, output=None, one_container=True, keep_container=False):
    check_project_dir(project)
    check_image(project)
    r = make_runner(project)
    with open(args) as fp:
        job = Job(wrapper, args=json.load(fp))
    try:
        result = Adun(r).execute(job, one_container, remove=not keep_container)
        if output:
            with open(output, 'w') as fp:
                to_json(result, fp)
        else:
            print(to_json(result))
    finally:
        c = r.chown()
        c.remove()



from sbdk.error import SbdkError


def check_project_dir(project):
    if not project.project_dir:
        raise SbdkError("Can't execute command: can't find project directory")


def check_project_data(project):
    if not project.project_data:
        raise SbdkError("Can't execute command: doesn't have a project data directory")


def check_project_state(project):
    if not project.state:
        raise SbdkError("Can't execute command: doesn't have a project state")


def check_image(project):
    if not project.image_id:
        raise SbdkError("Can't execute command: image-id not provided")



import os
import logging

from sbdk.commands import check_image
from sbdk.docker import make_runner
from sbdk.error import SbdkError


def argument_parser(subparsers):
    parser = subparsers.add_parser("fix-uid", help="Sets UID of wrapper user inside image to match your host UID")
    parser.set_defaults(func=fix_uid)

    parser.add_argument("--uid", "-u", type=str, default=str(os.getuid()),
                        help="Set this UID instead of current user's uid")


def fix_uid(project, uid=None):
    check_image(project)
    r = make_runner(project)
    c = r.run(['/sb/utils/fix-uid.sh', project.docker_user, uid])
    if not c.is_success():
        c.print_log()
        raise SbdkError("Error while fixing UID.")

    c.commit(message="fix-uid")
    if project.state:
        c.save_as_base()
    else:
        project.context['image_id'] = c.image['Id']
    logging.info("UID for user '%s' set to %s", project.docker_user, uid)
    c.remove()
    return c.image['Id']




import os


def insert_suffix(file_name, suffix):
    """
    Add a suffix before extension.

    >>> insert_suffix('file.txt', 'trimmed')
    'file.trimmed.txt'
    >>> insert_suffix('/path/to/file.trimmed.txt', 'sorted')
    'file.trimmed.sorted.txt'
    """
    name = os.path.basename(file_name)
    base, ext = os.path.splitext(name)
    return ''.join([base, '.', suffix, ext])


def change_ext(file_name, ext):
    """
    Change file extension.
    >>> change_ext('file.txt', 'csv')
    'file.csv'
    >>> change_ext('/path/to/file', 'txt')
    'file.txt'
    """
    name = os.path.basename(file_name)
    base, old_ext = os.path.splitext(name)
    return ''.join([base, '.', ext])




import json
import keyword
import os
import re
import stat
import random
import time
from os.path import join, isdir, dirname, basename, abspath, isfile

import sbdk.resources as resources
from sbdk.state import State
from sbdk.error import SbdkError
from sbdk import VERSION


DEFAULT_CONFIG = {
    'version': VERSION,
    'default_base_image': 'sevenbridges/sdkbase',
    'logging_config': 'logging.json',

    'shell_type': 'bash',
    'pip_options': '-q --download-cache /sb/pip_cache',
    'sdk_runner_command': ['/usr/bin/python', '-m', 'sbgsdk.cli'],
    'schema_server_url': 'https://hatchery.sbgenomics.com',
    'file_server_url': 'fs.sbgenomics.com',
    'vs_client_path': '/opt/vsclient/bin/vsclient.sh',

    'docker_registry_name': 'images.sbgenomics.com',
    'docker_registry_version': '1',
    'docker_registry_schema': 'https',
    'docker_daemon_url': 'unix://var/run/docker.sock',
    'docker_protocol_version': '1.8',
    'docker_user': 'sbg',
    'docker_dns': None,
    'docker_entrypoint': ['/sbin/my_init', '--quiet', '--']  # '--enable-insecure-key',
}


class Project(object):
    def __init__(self, context=None):
        context = context or {}
        self.sb_home = os.path.expanduser('~/.sbg')
        self._prepare_home_dir()

        pd = context.get('project_dir')
        if pd:
            pd = abspath(pd)
            if isdir(pd):
                self.project_dir = pd
            else:
                raise SbdkError("Directory does not exist: " + pd)
        else:
            cwd = context.get('cwd', '.')
            self.project_dir = find_project_dir(cwd)

        if not self.project_dir and context.get('is_test'):
            self.project_dir = '.'

        self.project_data = None
        self.state = None

        if self.project_dir:
            project_data = join(self.project_dir, '.sbdk')
            if isdir(project_data):
                self.project_data = project_data
                self.state = State(self.project_data)

        self.context = self._create_context(context)

    def _get_image_id(self):
        if self.context.get('image_id'):
            return self.context.get('image_id')
        elif self.state:
            return self.state.get('image_id')
        return None

    def _set_image_id(self, val):
        self.context['image_id'] = val

    image_id = property(_get_image_id, _set_image_id)

    def initialize(self, path, base_image=None):
        self.project_dir = abspath(destination(self.context.get('cwd', '.'), path))
        self.project_data = join(self.project_dir, '.sbdk')
        base_image = base_image or self.default_base_image

        if not isdir(self.project_data):
            os.makedirs(self.project_data)

        base = basename(self.project_dir)
        self.state = State(self.project_data)
        self.state['project_name'] = sanitize_project_name(base)
        self.state['image_id'] = base_image
        self.state.save()

    @property
    def base_cmd(self):
        log_conf = None
        if self.project_data:
            log_conf = join(self.project_data, self.context['logging_config'])
        if log_conf and isfile(log_conf):
            return self.context['sdk_runner_command'] + ['--logging-config',
                                                         join('.sbdk', self.context['logging_config'])]
        return self.context['sdk_runner_command']

    @property
    def name(self):
        name = None
        if self.state:
            name = self.state.get('project_name', None)
        if (not name) and self.project_dir:
            name = sanitize_project_name(basename(self.project_dir))
            if self.state:
                self.state['project_name'] = name
                self.state.save()
        return name

    @property
    def docker_registry_url(self):
        return '{0}://{1}/v{2}/'.format(self.docker_registry_schema,
                                        self.docker_registry_name,
                                        self.docker_registry_version)

    @property
    def repo_name(self):
        return self.docker_registry_name + "/" + self.state['username'] + "/" + self.name

    def __getattr__(self, item):
        if item in DEFAULT_CONFIG:
            return self.context[item]
        else:
            raise AttributeError("%r object has no attribute %r" %
                                (self.__class__.__name__, item))

    def _prepare_home_dir(self, retries=5):
        try:
            if not os.path.exists(self.sb_home):
                os.mkdir(self.sb_home)

            cache_path = join(self.sb_home, 'pip_cache')
            utils_path = join(self.sb_home, 'utils')

            for p in cache_path, utils_path:
                if not os.path.exists(p):
                    os.mkdir(p)

            resources.copy('fix-uid.sh', utils_path)
            fix_uid = join(utils_path, 'fix-uid.sh')
            mod = os.stat(fix_uid).st_mode
            plus_x = stat.S_IXGRP | stat.S_IXUSR | stat.S_IXOTH
            os.chmod(fix_uid, mod | plus_x)
        except OSError:
            if retries:
                time.sleep(random.random())
                return self._prepare_home_dir(retries-1)
            raise

    def _create_context(self, context=None):
        result = dict(DEFAULT_CONFIG)
        global_settings = join(self.sb_home, 'settings.json')
        if isfile(global_settings):
            with open(global_settings) as fp:
                result.update(json.load(fp))
        if self.project_data:
            project_settings = join(self.project_data, 'settings.json')
            if isfile(project_settings):
                with open(project_settings) as fp:
                    result.update(json.load(fp))
        result.update(context or {})
        return result


def destination(cwd, path):
    dst = os.path.join(cwd, path).rstrip('/')
    if os.path.isdir(dst) and os.path.exists(dst + '/.sbdk'):
        raise SbdkError("Project already initialized in dir '%s'" % dst)
    return dst


def find_project_dir(start_dir='.'):
    cd = abspath(start_dir)
    project_dir = None
    while cd != '/':
        if isdir(join(cd, '.sbdk')):
            project_dir = cd
            break
        cd = dirname(cd)
    return project_dir


def sanitize_project_name(name):
    project_name = name.replace('-', '_').lower()

    valid = (not keyword.iskeyword(project_name)) and \
        re.match('^[a-z][a-z0-9_]*$', project_name)

    if not valid:
        raise SbdkError("Invalid project name. Project name must be a valid Python identifier")
    return project_name

